import 'dart:async';
import 'dart:convert';
import 'dart:io';

import 'package:flutter/cupertino.dart';
import 'package:flutter_cache_store/flutter_cache_store.dart';
import 'package:http/http.dart' as http;
import 'package:http/http.dart';
import 'package:logger/logger.dart';

import 'config.dart';
import 'learning.dart';
import 'local/IFileManager.dart';
import 'local/local_file.dart';
import 'model/error_model.dart';
import 'model/header.dart';
import 'model/network_enum.dart';
import 'model/response_model.dart';
import 'serializable.dart';

class GenericRequestObject<ResponseType extends Serializable> {
  List<KeyValue> _headers;
  MethodType _methodType;
  Uri _uri;
  NetworkLearning _learning;
  NetworkConfig _config;
  bool _isLogging;
  dynamic body;
  ResponseType _type;
  List<Cookie> _cookies;
  Duration duration;
  IFileManager _fileManager;
  Map<String, String> _mapHeader;

  GenericRequestObject(this._methodType, this._learning, this._config,
      {this.duration, this.body}) {
    _headers = [];
    _cookies = [];
    _isLogging = false;
    _fileManager = _config?.fileManager ?? LocalFile();
    _mapHeader = {};
    _config?.headers?.forEach((element) {
      addHeader(element);
    });
  }

  Future<GenericRequestObject<ResponseType>> clearAllLocalData() async {
    await _fileManager.removeUserRequestCache(urlKeyOnLocalData);
    return this;
  }

  GenericRequestObject<ResponseType> path(String url) {
    _uri = Uri.parse(_config != null ? _config.baseUrl + url : url);
    return this;
  }

  GenericRequestObject<ResponseType> addHeader(KeyValue header) {
    if (header != null) {
      if (_mapHeader.containsKey(header.key)) {
        _mapHeader.update(header.key, (value) => header.value);
      } else {
        _mapHeader[header.key] = header.value;
      }
    }
    return this;
  }

  GenericRequestObject<ResponseType> addHeaders(List<KeyValue> header) {
    header?.forEach((element) => this.addHeader(element));
    return this;
  }

  GenericRequestObject<ResponseType> addCookie(Cookie cookie) {
    if (cookie != null) _cookies.add(cookie);
    return this;
  }

  GenericRequestObject<ResponseType> addHeaderWithParameters(
      {@required String key, @required String value}) {
    _headers.add(KeyValue(key, value));
    return this;
  }

  GenericRequestObject<ResponseType> responseType(ResponseType type) {
    _type = type;
    return this;
  }

  GenericRequestObject<ResponseType> enableLog(bool val) {
    _isLogging = val ?? false;
    return this;
  }

  GenericRequestObject<ResponseType> localManager(IFileManager manager) {
    _fileManager = manager;
    return this;
  }

  GenericRequestObject<ResponseType> query(List<KeyValue> querys) {
    querys?.forEach((element) {
      var old = _uri.toString();
      var prefix = old.contains("?") ? "&" : "?";
      _uri = Uri.parse("$old$prefix${element.key}=${element.value}");
    });

    return this;
  }

  void updateCookie(http.Response response) {
    String rawCookie = response.headers['set-cookie'];
    if (rawCookie != null) {
      int index = rawCookie.indexOf(';');
      final value = (index == -1) ? rawCookie : rawCookie.substring(0, index);
      _cookies.add(Cookie("cookie", value));
    }
  }

  Future<Response> _request({dynamic body}) async {
    switch (_methodType) {
      case MethodType.GET:
        return await http.get(_uri, headers: _mapHeader);
      case MethodType.POST:
        return await http.post(_uri, body: body, headers: _mapHeader);
      case MethodType.PUT:
        return await http.put(_uri, body: body, headers: _mapHeader);
      case MethodType.DELETE:
        return await http.delete(_uri, headers: _mapHeader);
      case MethodType.UPDATE:
        return await http.patch(_uri, body: body, headers: _mapHeader);
      default:
        throw ("Unknown method type");
    }
  }

  String get urlKeyOnLocalData =>
      "${_uri.toString()}-${_methodType.toString()}";

  Future<dynamic> fetch() async {
    Response response;
    dynamic body;
    ResponseModel model = ResponseModel();
    try {
      final localData = await getLocalData();

      if (localData != null && localData is String) {
        final jsonLocalData = jsonDecode(localData);
        model = saveBody(jsonLocalData, model);
        model.result = localData;
        return sendResponseModel(model);
      } else {
        response = await _request(body: body);
        updateCookie(response);
        if (response.statusCode >= HttpStatus.ok &&
            response.statusCode < HttpStatus.multipleChoices) {
          model.result = response.body;
          model.url = response.request.url.path;
          if (response.body != null && response.body.isNotEmpty) {
            final responseBody = jsonDecode(response.body);
            model = saveBody(responseBody, model);
            await writeCache(response.body);
          }
          return sendResponseModel(model);
        } else {
          ErrorModel error = ErrorModel();
          error.description = response.reasonPhrase;
          error.statusCode = response.statusCode;
          error.raw = response.body;
          error.request = response.request;

          if (_learning != null)
            return await _learning.checkCustomError(error);
          else
            return error;
        }
      }
    } on SocketException {
      return customErrorHandler(response, NetworkErrorTypes.NETWORK_ERROR);
    } on TimeoutException {
      return customErrorHandler(response, NetworkErrorTypes.TIMEOUT_ERROR);
    } catch (e) {
      Logger().e(e);
      return customErrorHandler(response, NetworkErrorTypes.CLIENT_ERROR);
    }
  }

  Future writeCache(String body) async {
    if (duration != null) {
      await _fileManager.writeUserRequestDataWithTime(
          urlKeyOnLocalData, body, duration);
    }
  }

  dynamic sendResponseModel(ResponseModel model) {
    if (_learning != null)
      return _learning?.checkSuccess<ResponseType>(model);
    else
      return model;
  }

  /// Parse [ResponseModel] on  [ResponseType]
  ///
  /// Response json is list, convert to [List<ResponseType>]
  /// Response json is map like a single object, convert to [ResponseType]
  /// finaly Response json is primitive type (ex. "1","veli",true), this stition not need convert return self [ResponseModel]
  ResponseModel saveBody(dynamic responseBody, ResponseModel model) {
    if (_isLogging) Logger().i(responseBody);

    if (responseBody is List) {
      model.data = responseBody
          .map((data) => _type.fromJson(data))
          .cast<ResponseType>()
          .toList();
    } else if (responseBody is Map) {
      model.data = _type.fromJson(responseBody) as ResponseType;
    } else {
      model.data = responseBody;
    }
    return model;
  }

  dynamic customErrorHandler(Response response, NetworkErrorTypes types) {
    ErrorModel error = ErrorModel();
    error.description = response?.body;
    error.type = types;

    if (_learning != null)
      return _learning.checkCustomError(error);
    else
      return error;
  }

  Future<String> getLocalData() async {
    final data =
        await _fileManager.getUserRequestDataOnString(urlKeyOnLocalData);
    return data;
  }
}
